#ifndef __VALHALLA_THOR_SERVICE_H__
#define __VALHALLA_THOR_SERVICE_H__

#include <cstdint>
#include <tuple>
#include <vector>

#include <boost/property_tree/ptree.hpp>

#include <valhalla/baldr/directededge.h>
#include <valhalla/baldr/graphid.h>
#include <valhalla/baldr/graphreader.h>
#include <valhalla/baldr/graphtile.h>
#include <valhalla/baldr/location.h>
#include <valhalla/meili/map_matcher_factory.h>
#include <valhalla/proto/directions_options.pb.h>
#include <valhalla/proto/trip.pb.h>
#include <valhalla/sif/costfactory.h>
#include <valhalla/sif/edgelabel.h>
#include <valhalla/thor/astar.h>
#include <valhalla/thor/attributes_controller.h>
#include <valhalla/thor/bidirectional_astar.h>
#include <valhalla/thor/isochrone.h>
#include <valhalla/thor/match_result.h>
#include <valhalla/thor/multimodal.h>
#include <valhalla/thor/timedep.h>
#include <valhalla/thor/triplegbuilder.h>
#include <valhalla/tyr/actor.h>
#include <valhalla/worker.h>

namespace valhalla {
namespace thor {

#ifdef HAVE_HTTP
void run_service(const boost::property_tree::ptree& config);
#endif

class thor_worker_t : public service_worker_t {
public:
  enum SOURCE_TO_TARGET_ALGORITHM { SELECT_OPTIMAL = 0, COST_MATRIX = 1, TIME_DISTANCE_MATRIX = 2 };
  thor_worker_t(const boost::property_tree::ptree& config,
                const std::shared_ptr<baldr::GraphReader>& graph_reader = {});
  virtual ~thor_worker_t();
#ifdef HAVE_HTTP
  virtual prime_server::worker_t::result_t work(const std::list<zmq::message_t>& job,
                                                void* request_info,
                                                const std::function<void()>& interrupt) override;
#endif
  virtual void cleanup() override;

  std::list<TripLeg> route(valhalla_request_t& request);
  std::string matrix(valhalla_request_t& request);
  std::list<TripLeg> optimized_route(valhalla_request_t& request);
  std::string isochrones(valhalla_request_t& request);
  std::list<TripLeg> trace_route(valhalla_request_t& request);
  std::string trace_attributes(valhalla_request_t& request);

protected:
  std::vector<thor::PathInfo> get_path(PathAlgorithm* path_algorithm,
                                       valhalla::Location& origin,
                                       valhalla::Location& destination,
                                       const std::string& costing);
  void log_admin(const TripLeg&);
  valhalla::sif::cost_ptr_t get_costing(const Costing costing, const DirectionsOptions& options);
  thor::PathAlgorithm* get_path_algorithm(const std::string& routetype,
                                          const valhalla::Location& origin,
                                          const valhalla::Location& destination);
  std::list<TripLeg> route_match(valhalla_request_t& request, const AttributesController& controller);
  std::vector<std::tuple<float, float, std::vector<thor::MatchResult>, std::list<TripLeg>>>
  map_match(valhalla_request_t& request,
            const AttributesController& controller,
            uint32_t best_paths = 1);
  TripLeg
  path_map_match(const std::vector<meili::MatchResult>& match_results,
                 const AttributesController& controller,
                 const std::vector<PathInfo>& path_edges,
                 std::unordered_map<size_t, std::pair<RouteDiscontinuity, RouteDiscontinuity>>&
                     route_discontinuities);
  std::list<TripLeg>
  path_arrive_by(google::protobuf::RepeatedPtrField<valhalla::Location>& correlated,
                 const std::string& costing);
  std::list<TripLeg>
  path_depart_at(google::protobuf::RepeatedPtrField<valhalla::Location>& correlated,
                 const std::string& costing);

  void parse_locations(valhalla_request_t& request);
  void parse_measurements(const valhalla_request_t& request);
  std::string parse_costing(const valhalla_request_t& request);
  void filter_attributes(const valhalla_request_t& request, AttributesController& controller);

  sif::TravelMode mode;
  std::vector<meili::Measurement> trace;
  sif::CostFactory<sif::DynamicCost> factory;
  sif::cost_ptr_t mode_costing[static_cast<int>(sif::TravelMode::kMaxTravelMode)];
  // Path algorithms (TODO - perhaps use a map?))
  AStarPathAlgorithm astar;
  BidirectionalAStar bidir_astar;
  MultiModalPathAlgorithm multi_modal_astar;
  TimeDepForward timedep_forward;
  TimeDepReverse timedep_reverse;
  Isochrone isochrone_gen;
  std::shared_ptr<meili::MapMatcher> matcher;
  float long_request;
  float max_timedep_distance;
  std::unordered_map<std::string, float> max_matrix_distance;
  SOURCE_TO_TARGET_ALGORITHM source_to_target_algorithm;
  meili::MapMatcherFactory matcher_factory;
  std::shared_ptr<baldr::GraphReader> reader;
};

} // namespace thor
} // namespace valhalla

#endif //__VALHALLA_THOR_SERVICE_H__
